@@ -1,5 +1,5 @@
 /* hdparm.c - Command line interface to get/set hard disk parameters */
-/*          - by Mark S. Lord (C) 1994-2001 -- freely distributable */
+/*          - by Mark Lord (C) 1994-2002 -- freely distributable */
 
 #include <unistd.h>
 #include <string.h>
@@ -16,13 +16,19 @@
 #include <sys/time.h>
 #include <sys/times.h>
 #include <sys/types.h>
+#include <sys/mount.h>
+#include <linux/types.h>
+#include <linux/fs.h>
 #include <linux/hdreg.h>
 #include <linux/major.h>
-#include <sys/mount.h>
 
-#define VERSION "v4.5"
+#include "hdparm.h"
 
-#undef	 DO_FLUSHCACHE		/* under construction: force cache flush on -W0 */
+extern const char *minor_str[];
+
+#define VERSION "v5.1"
+
+#undef DO_FLUSHCACHE		/* under construction: force cache flush on -W0 */
 
 #ifndef CDROM_SELECT_SPEED	/* already defined in 2.3.xx kernels and above */
 #define CDROM_SELECT_SPEED       0x5322
@@ -44,17 +50,19 @@
 #define BUFCACHE_FACTOR		2
 
 char *progname;
-static struct hd_driveid id;
 static int verbose = 0, get_identity = 0, get_geom = 0, noisy = 1, quiet = 0;
 static int flagcount = 0, do_flush = 0, is_scsi_hd = 0, is_xt_hd = 0;
 static int do_ctimings, do_timings = 0, old_identity = 0, timing_MB = TIMING_MB;
 
 static unsigned long set_readahead= 0, get_readahead= 0, readahead= 0;     
-static unsigned long set_readonly = 0, get_readonly = 0, readonly = 0;     
+static unsigned long set_readonly = 0, get_readonly = 0, readonly = 0;
 static unsigned long set_unmask   = 0, get_unmask   = 0, unmask   = 0;     
 static unsigned long set_mult     = 0, get_mult     = 0, mult     = 0;     
 #ifdef HDIO_SET_DMA
-static unsigned long set_dma      = 0, get_dma      = 0, dma      = 0;     
+static unsigned long set_dma      = 0, get_dma      = 0, dma      = 0;
+#endif
+#ifdef HDIO_GET_QDMA
+static unsigned long set_dma_q	  = 0, get_dma_q    = 0, dma_q	  = 0;
 #endif
 #ifdef HDIO_SET_NOWERR
 static unsigned long set_nowerr   = 0, get_nowerr   = 0, nowerr   = 0;
@@ -66,12 +74,14 @@ static unsigned long set_keep     = 0, get_keep     = 0, keep     = 0;
 static unsigned long set_io32bit  = 0, get_io32bit  = 0, io32bit  = 0;
 #endif
 #ifdef HDIO_SET_PIO_MODE
-static unsigned long set_piomode  = 0, noisy_piomode= 0, piomode  = 0;
+static unsigned long set_piomode  = 0, noisy_piomode= 0;
+int piomode = 0;
 #endif
 #ifdef HDIO_DRIVE_CMD
 static unsigned long set_dkeep    = 0, get_dkeep    = 0, dkeep    = 0;
-static unsigned long set_standby  = 0, get_standby  = 0, standby  = 0;
-static unsigned long set_xfermode = 0, get_xfermode = 0, xfermode = 0;
+static unsigned long set_standby  = 0, get_standby  = 0, standby_requested= 0;
+static unsigned long set_xfermode = 0, get_xfermode = 0;
+static int xfermode_requested= 0;
 static unsigned long set_lookahead= 0, get_lookahead= 0, lookahead= 0;
 static unsigned long set_prefetch = 0, get_prefetch = 0, prefetch = 0;
 static unsigned long set_defects  = 0, get_defects  = 0, defects  = 0;
@@ -102,8 +112,10 @@ static int	hwif_irq = 0;
 #ifdef HDIO_GET_BUSSTATE
 static int	set_busstate = 0, get_busstate = 0, busstate = 0;
 #endif
-
-extern void identify (void *, const char *);
+static int	reread_partn = 0;
+#ifdef HDIO_GET_ACOUSTIC
+static int	set_acoustic = 0, get_acoustic = 0, acoustic = 0;
+#endif
 
 #ifdef HDIO_DRIVE_RESET
 static int	perform_reset = 0;
@@ -112,61 +124,19 @@ static int	perform_reset = 0;
 static int	perform_tristate = 0,	tristate = 0;
 #endif /* HDIO_TRISTATE_HWIF */
 
-char *cfg_str[] =
-{	"",             " HardSect",   " SoftSect",  " NotMFM",
+const char *cfg_str[] =
+{	"",	     " HardSect",   " SoftSect",  " NotMFM",
 	" HdSw>15uSec", " SpinMotCtl", " Fixed",     " Removeable",
 	" DTR<=5Mbs",   " DTR>5Mbs",   " DTR>10Mbs", " RotSpdTol>.5%",
 	" dStbOff",     " TrkOff",     " FmtGapReq", " nonMagnetic"
 };
 
-char *ata_major_str[] = {
-	"reserved ",	"ATA-1 ",	"ATA-2 ",	"ATA-3 ",
-	"ATA-4 ",	"ATA-5 ",	"ATA-6 ",	"ATA-7 ",
-	"ATA-8 ",	"ATA-9 ",	"ATA-10 ",	"ATA-11 ",
-	"ATA-12 ",	"ATA-13 ",	"ATA-14 ",	"reserved "
-};
-
-char *ata_minor_str[] = {
-	"Reserved : ",						/* 0x0000 */
-	"ATA (ATA-1) X3T9.2 781D prior to revision 4 : ",	/* 0x0001 */
-	"ATA-1 published, ANSI X3.221-1994 : ",			/* 0x0002 */
-	"ATA (ATA-1) X3T10 781D revision 4 : ",			/* 0x0003 */
-	"ATA-2 published, ANSI X3.279-1996 : ",			/* 0x0004 */
-	"ATA-2 X3T10 948D prior to revision 2k : ",		/* 0x0005 */
-	"ATA-3 X3T10 2008D revision 1 : ",			/* 0x0006 */
-	"ATA-2 X3T10 948D revision 2k : ",			/* 0x0007 */
-	"ATA-3 X3T10 2008D revision 0 : ",			/* 0x0008 */
-	"ATA-2 X3T10 948D revision 3 : ",			/* 0x0009 */
-	"ATA-3 published, ANSI X3.298-199x : ",			/* 0x000A */
-	"ATA-3 X3T10 2008D revision 6 : ",			/* 0x000B */
-	"ATA-3 X3T13 2008D revision 7 and 7a : ",		/* 0x000C */
-	"ATA/ATAPI-4 X3T13 1153D revision 6 : ",		/* 0x000D */
-	"ATA/ATAPI-4 T13 1153D revision 13 : ",			/* 0x000E */
-	"ATA/ATAPI-4 X3T13 1153D revision 7 : ",		/* 0x000F */
-	"ATA/ATAPI-4 T13 1153D revision 18 : ",			/* 0x0010 */
-	"ATA/ATAPI-4 T13 1153D revision 15 : ",			/* 0x0011 */
-	"ATA/ATAPI-4 published, ANSI NCITS 317-1998 : ",	/* 0x0012 */
-	"ATA/ATAPI-5 T13 1321D revision 3 : ",			/* 0x0013 */
-	"ATA/ATAPI-4 T13 1153D revision 14 : ",			/* 0x0014 */
-	"ATA/ATAPI-5 T13 1321D revision 1 : ",			/* 0x0015 */
-	"Reserved : ",						/* 0x0016 */
-	"ATA/ATAPI-4 T13 1153D revision 17 : ",			/* 0x0017 */
-	"ATA/ATAPI-6 T13 1410D revision 0 : ",			/* 0x0018 */
-	"Reserved : "						/* 0x0019 */
-	"Reserved : "						/* 0x001A */
-	"Reserved : "						/* 0x001B */
-	"ATA/ATAPI-6 T13 1410D revision 1 : "			/* 0x001C */
-	"Reserved : "						/* 0x001D */
-	"Reserved : "						/* 0x001E */
-	"Reserved : "						/* 0x001F */
-};
-
-char *SlowMedFast[]	= {"slow", "medium", "fast", "eide", "ata"};
-char *BuffType[]	= {"unknown", "1Sect", "DualPort", "DualPortCache"};
+const char *SlowMedFast[]	= {"slow", "medium", "fast", "eide", "ata"};
+const char *BuffType[]	= {"unknown", "1Sect", "DualPort", "DualPortCache"};
 
 #define YN(b)	(((b)==0)?"no":"yes")
 
-static void dmpstr (char *prefix, unsigned int i, char *s[], unsigned int maxi)
+static void dmpstr (const char *prefix, unsigned int i, const char *s[], unsigned int maxi)
 {
 	if (i > maxi)
 		printf("%s%u", prefix, i);
@@ -174,13 +144,13 @@ static void dmpstr (char *prefix, unsigned int i, char *s[], unsigned int maxi)
 		printf("%s%s", prefix, s[i]);
 }
 
-static void dump_identity (struct hd_driveid *id)
+static void dump_identity (const struct hd_driveid *id)
 {
 	int i;
-	char pmodes[64] = {0,}, dmodes[128]={0,};
-	unsigned short int *id_regs=(void*)id;
+	char pmodes[64] = {0,}, dmodes[128]={0,}, umodes[128]={0,};
+	const unsigned short int *id_regs= (const void*) id;
 
-	printf("\n Model=%.40s, FwRev=%.8s, SerialNo=%.20s", 
+	printf("\n Model=%.40s, FwRev=%.8s, SerialNo=%.20s",
 		id->model, id->fw_rev, id->serial_no);
 	printf("\n Config={");
 	for (i=0; i<=15; i++) {
@@ -191,7 +161,7 @@ static void dump_identity (struct hd_driveid *id)
 	printf(" RawCHS=%u/%u/%u, TrkSize=%u, SectSize=%u, ECCbytes=%u\n",
 		id->cyls, id->heads, id->sectors,
 		id->track_bytes, id->sector_bytes, id->ecc_bytes);
-	dmpstr (" BuffType=",id->buf_type,BuffType,3);
+	dmpstr(" BuffType=",id->buf_type,BuffType,3);
 	printf(", BuffSize=%ukB, MaxMultSect=%u", id->buf_size/2, id->max_multsect);
 	if (id->max_multsect) {
 		printf(", MultSect=");
@@ -262,27 +232,27 @@ static void dump_identity (struct hd_driveid *id)
 			if (old_identity) {
 				strcat(dmodes," udma??");
 			} else {
-				if (id->dma_ultra & 0x100)	strcat(dmodes,"*");
-				if (id->dma_ultra & 0x001)	strcat(dmodes,"udma0 ");
-				if (id->dma_ultra & 0x200)	strcat(dmodes,"*");
-				if (id->dma_ultra & 0x002)	strcat(dmodes,"udma1 ");
-				if (id->dma_ultra & 0x400)	strcat(dmodes,"*");
-				if (id->dma_ultra & 0x004)	strcat(dmodes,"udma2 ");
+				if (id->dma_ultra & 0x100)	strcat(umodes,"*");
+				if (id->dma_ultra & 0x001)	strcat(umodes,"udma0 ");
+				if (id->dma_ultra & 0x200)	strcat(umodes,"*");
+				if (id->dma_ultra & 0x002)	strcat(umodes,"udma1 ");
+				if (id->dma_ultra & 0x400)	strcat(umodes,"*");
+				if (id->dma_ultra & 0x004)	strcat(umodes,"udma2 ");
 #ifdef __NEW_HD_DRIVE_ID
 				if (id->hw_config & 0x2000) {
 #else /* !__NEW_HD_DRIVE_ID */
 				if (id->word93 & 0x2000) {
 #endif /* __NEW_HD_DRIVE_ID */
-					if (id->dma_ultra & 0x800)	strcat(dmodes,"*");
-					if (id->dma_ultra & 0x008)	strcat(dmodes,"udma3 ");
-					if (id->dma_ultra & 0x1000)	strcat(dmodes,"*");
-					if (id->dma_ultra & 0x0010)	strcat(dmodes,"udma4 ");
-					if (id->dma_ultra & 0x2000)	strcat(dmodes,"*");
-					if (id->dma_ultra & 0x0020)	strcat(dmodes,"udma5 ");
-					if (id->dma_ultra & 0x4000)	strcat(dmodes,"*");
-					if (id->dma_ultra & 0x0040)	strcat(dmodes,"udma6 ");
-					if (id->dma_ultra & 0x8000)	strcat(dmodes,"*");
-					if (id->dma_ultra & 0x0080)	strcat(dmodes,"udma7 ");
+					if (id->dma_ultra & 0x800)	strcat(umodes,"*");
+					if (id->dma_ultra & 0x008)	strcat(umodes,"udma3 ");
+					if (id->dma_ultra & 0x1000)	strcat(umodes,"*");
+					if (id->dma_ultra & 0x0010)	strcat(umodes,"udma4 ");
+					if (id->dma_ultra & 0x2000)	strcat(umodes,"*");
+					if (id->dma_ultra & 0x0020)	strcat(umodes,"udma5 ");
+					if (id->dma_ultra & 0x4000)	strcat(umodes,"*");
+					if (id->dma_ultra & 0x0040)	strcat(umodes,"udma6 ");
+					if (id->dma_ultra & 0x8000)	strcat(umodes,"*");
+					if (id->dma_ultra & 0x0080)	strcat(umodes,"udma7 ");
 				}
 			}
 #endif
@@ -290,8 +260,11 @@ static void dump_identity (struct hd_driveid *id)
 	}
 	if ((id->capability&1) && (id->field_valid&2))
 		printf(", tDMA={min:%u,rec:%u}", id->eide_dma_min, id->eide_dma_time);
-	printf("\n PIO modes: %s", pmodes);
-	printf("\n DMA modes: %s", dmodes);
+	printf("\n PIO modes:  %s", pmodes);
+	if (*dmodes)
+		printf("\n DMA modes:  %s", dmodes);
+	if (*umodes)
+		printf("\n UDMA modes: %s", umodes);
 
 	printf("\n AdvancedPM=%s",YN(id_regs[83]&8));
 	if (id_regs[83] & 8) { 
@@ -306,11 +279,14 @@ static void dump_identity (struct hd_driveid *id)
 		printf(" WriteCache=%s",(id_regs[85]&0x20) ? "enabled" : "disabled");
 #ifdef __NEW_HD_DRIVE_ID
 	if ((id->minor_rev_num) || (id->major_rev_num)) {
-		printf("\n Drive Supports : ");
-		printf("%s", ata_minor_str[id->minor_rev_num]);
+		printf("\n Drive conforms to: ");
+		if (id->minor_rev_num <= 31)
+			printf("%s: ", minor_str[id->minor_rev_num]);
+		else
+			printf("unknown: ");
 		for (i=0; i<=15; i++) {
 			if (id->major_rev_num & (1<<i))
-				printf("%s", ata_major_str[i]);
+				printf(" %u", i);
 		}
 	}
 #endif /* __NEW_HD_DRIVE_ID */
@@ -327,8 +303,8 @@ void flush_buffer_cache (int fd)
 	if (is_scsi_hd || is_xt_hd) {
 		sleep(1);
 	} else {
-		if (ioctl(fd, HDIO_DRIVE_CMD, NULL))	/* await completion */
-			perror("HDIO_DRIVE_CMD(null) failed");
+		if (ioctl(fd, HDIO_DRIVE_CMD, NULL) && errno != EINVAL)	/* await completion */
+			perror("HDIO_DRIVE_CMD(null) (wait for flush complete) failed");
 	}
 #endif
 }
@@ -346,10 +322,14 @@ int read_big_block (int fd, char *buf)
 {
 	int i, rc;
 	if ((rc = read(fd, buf, TIMING_BUF_BYTES)) != TIMING_BUF_BYTES) {
-		if (rc)
-			perror("read() failed");
-		else
+		if (rc) {
+			if (rc == -1)
+				perror("read() failed");
+			else
+				fprintf(stderr, "read(%u) returned %u bytes\n", TIMING_BUF_BYTES, rc);
+		} else {
 			fputs ("read() hit EOF - device too small\n", stderr);
+		}
 		return 1;
 	}
 	/* access all sectors of buf to ensure the read fully completed */
@@ -540,7 +520,7 @@ static void on_off (unsigned int value)
 #ifdef HDIO_GET_BUSSTATE
 static void bus_state_value (unsigned int value)
 {
-	char *string;
+	const char *string;
 
 	switch (value) {
 	case BUSSTATE_ON:
@@ -597,6 +577,66 @@ static void interpret_standby (unsigned int standby)
 	printf(")\n");
 }
 
+struct xfermode_entry {
+    int val;
+    const char *name;
+};
+
+static const struct xfermode_entry xfermode_table[] = {
+    { 8,    "pio0" },
+    { 9,    "pio1" },
+    { 10,   "pio2" },
+    { 11,   "pio3" },
+    { 12,   "pio4" },
+    { 13,   "pio5" },
+    { 14,   "pio6" },
+    { 15,   "pio7" },
+    { 16,   "sdma0" },
+    { 17,   "sdma1" },
+    { 18,   "sdma2" },
+    { 19,   "sdma3" },
+    { 20,   "sdma4" },
+    { 21,   "sdma5" },
+    { 22,   "sdma6" },
+    { 23,   "sdma7" },
+    { 32,   "mdma0" },
+    { 33,   "mdma1" },
+    { 34,   "mdma2" },
+    { 35,   "mdma3" },
+    { 36,   "mdma4" },
+    { 37,   "mdma5" },
+    { 38,   "mdma6" },
+    { 39,   "mdma7" },
+    { 64,   "udma0" },
+    { 65,   "udma1" },
+    { 66,   "udma2" },
+    { 67,   "udma3" },
+    { 68,   "udma4" },
+    { 69,   "udma5" },
+    { 70,   "udma6" },
+    { 71,   "udma7" },
+    { 0, NULL }
+};
+
+static int translate_xfermode(char * name)
+{
+    const struct xfermode_entry *tmp;
+    char *endptr;
+    int val = -1;
+
+
+    for (tmp = xfermode_table; tmp->name != NULL; ++tmp) {
+	if (!strcmp(name, tmp->name))
+	    return tmp->val;
+    }
+
+    val = strtol(name, &endptr, 10);
+    if (*endptr == '\0')
+	return val;
+
+    return -1;
+}
+
 static void interpret_xfermode (unsigned int xfermode)
 {
 	printf(" (");
@@ -766,7 +806,7 @@ void process_dev (char *devname)
 			if (piomode == 255)
 				printf(" attempting to auto-tune PIO mode\n");
 			else
-				printf(" attempting to set PIO mode to %ld\n", piomode);
+				printf(" attempting to set PIO mode to %d\n", piomode);
 		}
 		if (ioctl(fd, HDIO_SET_PIO_MODE, piomode)) 
 			perror(" HDIO_SET_PIO_MODE failed");
@@ -777,7 +817,7 @@ void process_dev (char *devname)
 		no_scsi();
 		no_xt();
 		if (get_io32bit)
-			printf(" setting 32-bit I/O support flag to %ld\n", io32bit);
+			printf(" setting 32-bit IO_support flag to %ld\n", io32bit);
 		if (ioctl(fd, HDIO_SET_32BIT, io32bit)) 
 			perror(" HDIO_SET_32BIT failed");
 	}
@@ -822,6 +862,17 @@ void process_dev (char *devname)
 			perror(" HDIO_SET_DMA failed");
 	}
 #endif
+#ifdef HDIO_SET_QDMA
+	if (set_dma_q) {
+		no_scsi();
+		if (get_dma_q) {
+			printf(" setting DMA queue_depth to %ld", dma_q);
+			on_off(dma_q);
+		}
+		if (ioctl(fd, HDIO_SET_QDMA, dma_q))
+			perror(" HDIO_SET_QDMA failed");
+	}
+#endif
 #ifdef HDIO_SET_NOWERR
 	if (set_nowerr) {
 		no_scsi();
@@ -895,10 +946,10 @@ void process_dev (char *devname)
 		unsigned char args[4] = {WIN_SETFEATURES,0,3,0};
 		no_scsi();
 		no_xt();
-		args[1] = xfermode;
+		args[1] = xfermode_requested;
 		if (get_xfermode) {
-			printf(" setting xfermode to %ld", xfermode);
-			interpret_xfermode(xfermode);
+			printf(" setting xfermode to %d", xfermode_requested);
+			interpret_xfermode(xfermode_requested);
 		}
 		if (ioctl(fd, HDIO_DRIVE_CMD, &args))
 			perror(" HDIO_DRIVE_CMD(setxfermode) failed");
@@ -922,18 +973,18 @@ void process_dev (char *devname)
 		if (apmmode>255) apmmode=255;
 		if (get_apmmode) 
 			printf(" setting Advanced Power Management level to");
-			if (apmmode==255) { 
+		if (apmmode==255) { 
 			/* disable Advanced Power Management */
-				args[2] = 0x85; /* feature register */ 
-				if (get_apmmode) printf(" disabled\n");
-			} else {
-				/* set Advanced Power Management mode */
-				args[2] = 0x05; /* feature register */  
-				args[1] = apmmode; /* sector count register */
-				if (get_apmmode) printf(" 0x%02lX (%ld)\n",apmmode,apmmode);
-			} 
-			if (ioctl(fd, HDIO_DRIVE_CMD, &args))
-				perror(" HDIO_DRIVE_CMD failed");
+			args[2] = 0x85; /* feature register */ 
+			if (get_apmmode) printf(" disabled\n");
+		} else {
+			/* set Advanced Power Management mode */
+			args[2] = 0x05; /* feature register */  
+			args[1] = apmmode; /* sector count register */
+			if (get_apmmode) printf(" 0x%02lX (%ld)\n",apmmode,apmmode);
+		} 
+		if (ioctl(fd, HDIO_DRIVE_CMD, &args))
+			perror(" HDIO_DRIVE_CMD failed");
 	}
 #ifdef CDROM_SELECT_SPEED
 	if (set_cdromspeed) {
@@ -941,7 +992,7 @@ void process_dev (char *devname)
 		if (ioctl (fd, CDROM_SELECT_SPEED, cdromspeed))
 			perror(" CDROM_SELECT_SPEED failed");
 	}
-#endif                                    
+#endif
 	if (set_wcache) {
 #ifdef DO_FLUSHCACHE
 #ifndef WIN_FLUSHCACHE
@@ -1010,12 +1061,12 @@ void process_dev (char *devname)
 			perror(" HDIO_DRIVE_CMD(seagatepwrsave) failed");
 	}
 	if (set_standby) {
-		unsigned char args[4] = {WIN_SETIDLE1,standby,0,0};
+		unsigned char args[4] = {WIN_SETIDLE1,standby_requested,0,0};
 		no_scsi();
 		no_xt();
 		if (get_standby) {
-			printf(" setting standby to %ld", standby);
-			interpret_standby(standby);
+			printf(" setting standby to %lu", standby_requested);
+			interpret_standby(standby_requested);
 		}
 		if (ioctl(fd, HDIO_DRIVE_CMD, &args))
 			perror(" HDIO_DRIVE_CMD(setidle1) failed");
@@ -1050,15 +1101,17 @@ void process_dev (char *devname)
 		if (ioctl(fd, HDIO_GET_32BIT, &parm))
 			perror(" HDIO_GET_32BIT failed");
 		else {
-			printf(" I/O support  =%3ld (", parm);
+			printf(" IO_support   =%3ld (", parm);
 			switch (parm) {
 				case 0:	printf("default ");
 				case 2: printf("16-bit)\n");
 					break;
 				case 1:	printf("32-bit)\n");
-					break; 
+					break;
 				case 3:	printf("32-bit w/sync)\n");
-					break; 
+					break;
+				case 8:	printf("Request-Queue-Bypass)\n");
+					break;
 				default:printf("\?\?\?)\n");
 			}
 		}
@@ -1074,6 +1127,7 @@ void process_dev (char *devname)
 			on_off(parm);
 		}
 	}
+
 #ifdef HDIO_GET_DMA
 	if ((verbose && !is_scsi_hd) || get_dma) {
 		no_scsi();
@@ -1081,6 +1135,20 @@ void process_dev (char *devname)
 			perror(" HDIO_GET_DMA failed");
 		else {
 			printf(" using_dma    = %2ld", parm);
+			if (parm == 8)
+				printf(" (DMA-Assisted-PIO)\n");
+			else
+				on_off(parm);
+		}
+	}
+#endif
+#ifdef HDIO_GET_QDMA 
+	if (get_dma_q) {
+		no_scsi();
+		if(ioctl(fd, HDIO_GET_QDMA, &parm)) {
+			perror(" HDIO_GET_QDMA failed");
+		} else {
+			printf(" queue_depth  = %2ld", parm);
 			on_off(parm);
 		}
 	}
@@ -1099,7 +1167,7 @@ void process_dev (char *devname)
 #endif /* HDIO_GET_KEEPSETTINGS */
 
 #ifdef HDIO_SET_NOWERR
-	if ((verbose && !is_scsi_hd && !is_xt_hd) || get_nowerr) {
+	if (get_nowerr) {
 		no_scsi();
 		no_xt();
 		if (ioctl(fd, HDIO_GET_NOWERR, &parm))
@@ -1163,7 +1231,7 @@ void process_dev (char *devname)
 				state = "sleeping";
 		} else {
 			state = (args[2] == 255) ? "active/idle" : "standby";
-		} 
+		}
 		printf(" drive state is:  %s\n", state);
 	}
 #endif
@@ -1172,7 +1240,7 @@ void process_dev (char *devname)
 		no_scsi();
 		no_xt();
 		if (ioctl(fd, HDIO_DRIVE_RESET, NULL))
-			 perror(" HDIO_DRIVE_RESET failed");
+			perror(" HDIO_DRIVE_RESET failed");
 	}
 #endif /* HDIO_DRIVE_RESET */
 #ifdef HDIO_TRISTATE_HWIF
@@ -1185,6 +1253,8 @@ void process_dev (char *devname)
 	}
 #endif /* HDIO_TRISTATE_HWIF */
 	if (get_identity) {
+		static struct hd_driveid id;
+
 		no_scsi();
 		no_xt();
 #ifdef HDIO_OBSOLETE_IDENTITY
@@ -1232,7 +1302,7 @@ void process_dev (char *devname)
 	}
 #endif
 #ifdef HDIO_GET_BUSSTATE
-	if ((verbose && !is_scsi_hd) || get_busstate) {
+	if (get_busstate) {
 		no_scsi();
 		if (ioctl(fd, HDIO_GET_BUSSTATE, &parm))
 			perror(" HDIO_GET_BUSSTATE failed");
@@ -1242,6 +1312,31 @@ void process_dev (char *devname)
 		}
 	}
 #endif
+	if (reread_partn) {
+		if (ioctl(fd, BLKRRPART, NULL)) {
+			perror(" BLKRRPART failed");
+		}
+	}
+ #ifdef HDIO_SET_ACOUSTIC
+	if (set_acoustic) {
+		no_scsi();
+		if (get_acoustic) {
+			printf(" setting acoustic management to %d\n", acoustic);
+		}
+		if (ioctl(fd, HDIO_SET_ACOUSTIC, acoustic))
+			perror(" HDIO_SET_ACOUSTIC failed");
+	}
+ #endif /* HDIO_SET_ACOUSTIC */
+ #ifdef HDIO_GET_ACOUSTIC
+	if (get_acoustic) {
+		no_scsi();
+		if (ioctl(fd, HDIO_GET_ACOUSTIC, &parm)) {
+			perror(" HDIO_GET_ACOUSTIC failed");
+		} else {
+			printf(" acoustic     = %2ld (128=quiet ... 254=fast)\n", parm);
+		}
+	}
+ #endif /* HDIO_GET_ACOUSTIC */
 	if (do_ctimings)
 		time_cache (fd);
 	if (do_timings)
@@ -1263,7 +1358,7 @@ void usage_error (void)
 #ifdef HDIO_GET_BUSSTATE
 	" -b   get/set bus state (0 == off, 1 == on, 2 == tristate)\n"
 #endif
-	" -B   get Advanced Power Management setting (1-255)\n"
+	" -B   set Advanced Power Management setting (1-255)\n"
 #ifdef HDIO_SET_32BIT
 	" -c   get/set IDE 32-bit IO setting\n"
 #endif
@@ -1283,6 +1378,7 @@ void usage_error (void)
 	" -i   display drive identification\n"
 #ifdef HDIO_SET_32BIT
 	" -I   detailed/current information directly from drive\n"
+	" -Istdin  similar to -I, but wants /proc/ide/*/hd?/identify as input\n"
 #endif
 #ifdef HDIO_GET_KEEPSETTINGS
 	" -k   get/set keep_settings_over_reset flag (0/1)\n"
@@ -1291,8 +1387,11 @@ void usage_error (void)
 	" -K   set drive keep_features_over_reset flag (0/1)\n"
 	" -L   set drive doorlock (0/1) (removable harddisks only)\n"
 #endif
+#ifdef HDIO_GET_ACOUSTIC
+	" -M   get/set acoustic management (0-254, 128: quiet, 254: fast) (EXPERIMENTAL)\n"
+#endif /* HDIO_GET_ACOUSTIC */
 	" -m   get/set multiple sector count\n"
-#ifdef HDIO_SET_NOWERR 
+#ifdef HDIO_SET_NOWERR
 	" -n   get/set ignore-write-errors flag (0/1)\n"
 #endif
 #ifdef HDIO_SET_PIO_MODE
@@ -1302,6 +1401,9 @@ void usage_error (void)
 	" -P   set drive prefetch count\n"
 #endif
 	" -q   change next setting quietly\n"
+#ifdef HDIO_GET_QDMA
+	" -Q   get/set DMA tagged-queuing depth (if supported)\n"
+#endif
 	" -r   get/set readonly flag (DANGEROUS to set)\n"
 #ifdef HDIO_SCAN_HWIF
 	" -R   register an IDE interface (DANGEROUS)\n"
@@ -1315,7 +1417,7 @@ void usage_error (void)
 #ifdef HDIO_UNREGISTER_HWIF
 	" -U   un-register an IDE interface (DANGEROUS)\n"
 #endif
-	" -v   default; same as -acdgkmnru (-gr for SCSI, -adgr for XT)\n"
+	" -v   defaults; same as -mcudkrag for IDE drives\n"
 	" -V   display program version and exit immediately\n"
 #ifdef HDIO_DRIVE_CMD
 #ifdef HDIO_DRIVE_RESET
@@ -1323,12 +1425,13 @@ void usage_error (void)
 #endif
 	" -W   set drive write-caching flag (0/1) (DANGEROUS)\n"
 #ifdef HDIO_TRISTATE_HWIF
-	" -x   perform device for hotswap flag (0/1) (DANGEROUS)\n"
+	" -x   tristate device for hotswap (0/1) (DANGEROUS)\n"
 #endif
 	" -X   set IDE xfer mode (DANGEROUS)\n"
 	" -y   put IDE drive in standby mode\n"
 	" -Y   put IDE drive to sleep\n"
 	" -Z   disable Seagate auto-powersaving mode\n"
+	" -z   re-read partition table\n"
 #endif
 	);
 	exit(1);
@@ -1342,19 +1445,70 @@ void usage_error (void)
 					num = (num * 10) + (*p++ - '0'); \
 				}
 
+#define GET_STRING(flag, num) tmpstr = name; \
+				tmpstr[0] = '\0'; \
+				if (!*p && argc && isalnum(**argv)) \
+					p = *argv++, --argc; \
+				while (isalnum(*p) && (tmpstr - name) < 31) { \
+					tmpstr[0] = *p++; \
+					tmpstr[1] = '\0'; \
+					++tmpstr; \
+				} \
+				num = translate_xfermode(name); \
+				if (num == -1) \
+					flag = 0; \
+				else \
+					flag = 1;
+
+static int fromhex (unsigned char c)
+{
+	if (c >= 'a' && c <= 'f')
+		return 10 + (c - 'a');
+	if (c >= '0' && c <= '9')
+		return (c - '0');
+	fprintf(stderr, "bad char: '%c' 0x%02x\n", c, c);
+	exit(-1);
+}
+
+static int identify_from_stdin (void)
+{
+	unsigned char *buf = malloc(1600), *b = buf;
+	unsigned short *w = (unsigned short *)buf;
+	int i, count = read(0, buf, 1280);
+
+	if (count != 1280) {
+		fprintf(stderr, "read(1280 bytes) failed (rc=%d)", count);
+		perror("");
+		exit(errno);
+	}
+	for (i = 0; count >= 4; ++i) {
+		*w++ = (fromhex(*b++)<<12) | (fromhex(*b++)<<8) | (fromhex(*b++)<<4) | (fromhex(*b++));
+		++b;
+		count -= 5;
+	}
+	identify((void *)buf, NULL);
+	return 0;
+}
+
 int main(int argc, char **argv)
 {
 	char c, *p;
+	char *tmpstr;
+	char name[32];
+	int neg;
 
 	if  ((progname = (char *) strrchr(*argv, '/')) == NULL)
-                progname = *argv;
-        else
-                progname++;
+		progname = *argv;
+	else
+		progname++;
 	++argv;
 
 	if (!--argc)
 		usage_error();
 	while (argc--) {
+		if (!strcmp("-Istdin", *argv)) {
+			exit(identify_from_stdin());
+		}
 		p = *argv++;
 		if (*p == '-') {
 			if (!*++p)
@@ -1403,7 +1557,7 @@ int main(int argc, char **argv)
 						noisy = 1;
 						if (!*p && argc && isdigit(**argv))
 							p = *argv++, --argc;
-						if (*p == '0' || *p == '1') {
+						if (*p >= '0' && *p <= '9') {
 							set_dma = 1;
 							dma = *p++ - '0';
 						}
@@ -1425,13 +1579,7 @@ int main(int argc, char **argv)
 					case 'p':
 						noisy_piomode = noisy;
 						noisy = 1;
-						if (!*p && argc && isdigit(**argv))
-							p = *argv++, --argc;
-						if (*p >= '0' && *p <= '9')
-							piomode = *p++ - '0';
-						else
-							piomode = 255;  /* auto-tune */
-						set_piomode = 1;
+						GET_STRING(set_piomode,piomode);
 						break;
 #endif
 					case 'r':
@@ -1460,7 +1608,7 @@ int main(int argc, char **argv)
 					case 'S':
 						get_standby = noisy;
 						noisy = 1;
-						GET_NUMBER(set_standby,standby);
+						GET_NUMBER(set_standby,standby_requested);
 						if (!set_standby)
 							fprintf(stderr, "-S: missing value\n");
 						break;
@@ -1491,7 +1639,7 @@ int main(int argc, char **argv)
 					case 'X':
 						get_xfermode = noisy;
 						noisy = 1;
-						GET_NUMBER(set_xfermode,xfermode);
+						GET_STRING(set_xfermode,xfermode_requested);
 						if (!set_xfermode)
 							fprintf(stderr, "-X: missing value\n");
 						break;
@@ -1561,6 +1709,10 @@ int main(int argc, char **argv)
 						set_sleepnow = 1;
 						break;
 
+					case 'z':
+						reread_partn = 1;
+						break;
+
 					case 'Z':
 						get_seagate = noisy;
 						noisy = 1;
@@ -1629,6 +1781,18 @@ int main(int argc, char **argv)
 						scan_hwif = 1;
 						break;
 #endif /* HDIO_SCAN_HWIF */
+#ifdef HDIO_GET_QDMA
+					case 'Q':
+						get_dma_q = noisy;
+						noisy = 1;
+#ifdef HDIO_SET_QDMA
+						neg = 0;
+						GET_NUMBER(set_dma_q, dma_q);
+						if (neg)
+							dma_q = -dma_q;
+#endif
+						break;
+#endif
 #ifdef HDIO_DRIVE_RESET
 					case 'w':
 						perform_reset = 1;
@@ -1652,6 +1816,7 @@ int main(int argc, char **argv)
 						GET_NUMBER(set_readahead,readahead);
 						break;
 					case 'B':
+						get_apmmode = noisy;
 						noisy = 1;
 						GET_NUMBER(set_apmmode,apmmode);
 						if (!set_apmmode)
@@ -1667,22 +1832,29 @@ int main(int argc, char **argv)
 						break;
 #ifdef HDIO_GET_BUSSTATE
 					case 'b':
-                                                get_busstate = noisy;
-                                                noisy = 1;
-                                                if (!*p && argc && isdigit(**argv))
-						  p = *argv++, --argc;
+						get_busstate = noisy;
+						noisy = 1;
+						if (!*p && argc && isdigit(**argv))
+							p = *argv++, --argc;
 						switch (*p) {
-						case '0':
-						case '1':
-						case '2':
-						  set_busstate = 1;
-						  busstate = *p++ - '0';
-						  break;
-						default:
-						  break;
-                                                }
+							case '0':
+							case '1':
+							case '2':
+								set_busstate = 1;
+								busstate = *p++ - '0';
+								break;
+							default:
+								break;
+						}
 						break;
 #endif
+#ifdef HDIO_GET_ACOUSTIC
+					case 'M':
+						get_acoustic = noisy; /* are we noisy? */
+						noisy = 1;
+						GET_NUMBER(set_acoustic,acoustic);
+						break;
+#endif /* HDIO_GET_ACOUSTIC */
 					case 'h':
 					default:
 						usage_error();
@@ -1692,7 +1864,7 @@ int main(int argc, char **argv)
 				usage_error();
 		} else {
 			process_dev (p);
-		} 
+		}
 	}
 	exit (0);
 }
